import os
import sys
import optparse
from unittest.main import main
import traci
import sumolib
from numpy import double
from sumolib import checkBinary 


"""
    @Class: Helper class with methods pertaining to stress calculation
"""

class StressHelper:

    """
        @Method: MinMax scaling of a dict parameter
    """
    @staticmethod
    def normalise(dict):
        xmax = -1 
        xmin = 100000
        for key in dict:
            xmax = max(xmax,dict[key])
            xmin = min(xmin,dict[key])

        if(xmax-xmin > 0):
            for key in dict: 
                dict[key] = (dict[key] - xmin)/(xmax-xmin) 

        return dict

    """
        @Method: Get specific stress parameter
    """
    @staticmethod
    def getStressAt(junction,parameter):
        value = 0 
        incoming_edges = junction.getIncoming()
        for edge in incoming_edges:
            lanes = edge.getLanes()
            for lane in lanes:
                lane_id = lane.getID()

                if(parameter == "haltingNumber"):
                    edgeValue = traci.lane.getLastStepHaltingNumber(lane_id)
                
                if(parameter == "CO2Emissions"):
                    edgeValue = traci.lane.getCO2Emission(lane_id)

                if(parameter == "StepOccupancy"):
                    edgeValue = traci.lane.getLastStepOccupancy(lane_id)
                
                if(parameter == "vehicleLength"):
                    edgeValue = traci.lane.getLastStepLength(lane_id)

                if(parameter == "waitingTime"):
                    edgeValue = traci.lane.getWaitingTime(lane_id)
                value += edgeValue
            
        return value

"""
    @Class: Main class for stress analysis
"""
class StressJunction:
    """
        @Method: Initialising variables and config
    """ 
    def __init__(self, maxTimeSteps, weightsArray, pathCFG, outPath, pathNET, pathSummaryFile, numLocs, initialPoiLocation = (5064.74,3568.48)):
        
        # Check for Bash Shortcut
        if "SUMO_HOME" in os.environ:
            sys.path.append(os.path.join(
                os.environ["SUMO_HOME"],
                "tools"
            ))
        else:
            sys.exit("Please populate $SUMO_HOME in your .bashrc file.") 

        # Assign all variables
        self.maxTimeSteps = maxTimeSteps    
        self.weightsArray = weightsArray
        self.pathCFG = pathCFG
        self.outPath = outPath
        self.numLocs = numLocs
        self.net = sumolib.net.readNet(pathNET)
        self.pathSummaryFile = pathSummaryFile
        self.initialPoiLocation = initialPoiLocation


        # Option Parser
        optParser = optparse.OptionParser()
        optParser.add_option(
            "--nogui", 
            action = "store_true",
            default = False,
            help = "Please run sumo --help."
        )
        options, args = optParser.parse_args()
        self.options = options

        # Initialise variables used in the simulation
        self.weights = {
            "haltingNumber":1,
            "CO2Emissions":1,
            "StepOccupancy":1,
            "vehicleLength":1,
            "waitingTime":1
        }

        if weightsArray[0] is not None:
            self.weights["haltingNumber"] = self.weightsArray[0]
        if weightsArray[1] is not None:
            self.weights["CO2Emissions"] = self.weightsArray[1]
        if weightsArray[2] is not None:
            self.weights["StepOccupancy"] = self.weightsArray[2]
        if weightsArray[3] is not None:
            self.weights["vehicleLength"] = self.weightsArray[3]
        if weightsArray[4] is not None:
            self.weights["waitingTime"] = self.weightsArray[4]
        
        self.junctions = self.net.getNodes()
        self.junctionsEmissions = []

        for _ in range(len(self.junctions)):
            self.junctionsEmissions.append([])

    """
        @Method: TraCI simulation step and calculation of positions - run in the runSimulation() method
    """
    def loop(self):
        traci.poi.add("index", self.initialPoiLocation[0], self.initialPoiLocation[1], (255,255,255), layer=202.0)
        step = 0

        steps = []
        stressedCounts = {}
        weights = self.weights
        max_list = []

        junctions = self.junctions
        locs = min(5, self.numLocs)
        

        while traci.simulation.getMinExpectedNumber() > 0:
            traci.simulationStep()
            traci.poi.setType('index', 'TimeStep = '+str(step))
            if step%100 == 0:
                if(step!=0):
                    for i in range(locs):
                        traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
                
                max_list = {}
                max_dict = {}
                parameterDict = {}
                for idx, junction in enumerate(junctions):
                    stress = StressHelper.getStressAt(junction,"haltingNumber")
                    self.junctionsEmissions[idx].append(stress)
                    parameterDict[idx] = stress
                max_dict["haltingNumber"] = parameterDict 

                parameterDict = {}
                for idx, junction in enumerate(junctions):
                    stress = StressHelper.getStressAt(junction,"CO2Emissions")
                    self.junctionsEmissions[idx].append(stress)
                    parameterDict[idx] = stress
                max_dict["CO2Emissions"] = parameterDict 

                parameterDict = {}
                for idx, junction in enumerate(junctions):
                    stress = StressHelper.getStressAt(junction,"StepOccupancy")
                    self.junctionsEmissions[idx].append(stress)
                    parameterDict[idx] = stress
                max_dict["StepOccupancy"] = parameterDict 

                parameterDict = {}
                for idx, junction in enumerate(junctions):
                    stress = StressHelper.getStressAt(junction,"vehicleLength")
                    self.junctionsEmissions[idx].append(stress)
                    parameterDict[idx] = stress
                max_dict["vehicleLength"] = parameterDict 

                parameterDict = {}
                for idx, junction in enumerate(junctions):
                    stress = StressHelper.getStressAt(junction,"waitingTime")
                    self.junctionsEmissions[idx].append(stress)
                    parameterDict[idx] = stress
                max_dict["waitingTime"] = parameterDict 

                #max dict is now having all values 
                #Now normalize 
                max_dict["haltingNumber"] = StressHelper.normalise(max_dict["haltingNumber"])
                max_dict["CO2Emissions"] = StressHelper.normalise(max_dict["CO2Emissions"])
                max_dict["StepOccupancy"] = StressHelper.normalise(max_dict["StepOccupancy"])
                max_dict["vehicleLength"] = StressHelper.normalise(max_dict["vehicleLength"])
                max_dict["waitingTime"] = StressHelper.normalise(max_dict["waitingTime"])

                for edge in max_dict["CO2Emissions"]:
                    max_list[edge] = 0
                
                for param in max_dict: 
                    dict = max_dict[param]
                    for edge in dict: 
                        max_list[edge] +=  weights[param] * dict[edge]
                        
                weighted_values = [] 
                for key in max_list: 
                    weighted_values.append((int(key),double(max_list[key]))) 

                    
                max_list = weighted_values
                max_list = sorted(max_list,key = lambda x:x[1])
                max_list.reverse()

                steps.append(step)

                for i in range(locs):
                    traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
                    if junctions[max_list[i][0]].getID() in stressedCounts.keys():
                        stressedCounts[junctions[max_list[i][0]].getID()]+=1
                    else:
                        stressedCounts[junctions[max_list[i][0]].getID()]=1
                
                filename = self.outPath+'stress/screenshots/00'+str(step)+'.jpg'
                traci.gui.screenshot('View #0', filename)
            
            step += 1

        # Printing the values found
        traci.poi.setType('index', 'Aggregate')
        max_stressed = sorted(stressedCounts.items(), key= lambda x:x[1])
        max_stressed.reverse()
        
        for i in range(locs):
            traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
        
        f = open(self.pathSummaryFile, 'w')
        for i in range(locs):
            traci.gui.toggleSelection(max_stressed[i][0], objType='junction')
            x,y = traci.junction.getPosition(max_stressed[i][0])
            lon, lat = traci.simulation.convertGeo(x, y)
            lon, lat = self.net.convertXY2LonLat(x, y)
            f.write(str(max_stressed[i][0])+" Coordinates:" + str(lat) + " " + str(lon) + '\n')
        f.close()

        filename = self.outPath+'stress/screenshots/aggregate.jpg'
        traci.simulationStep()
        traci.gui.screenshot('View #0', filename)   
        traci.simulationStep()
        traci.close()

    """
        @Method: Main Method for Speed Camera simulation
    """
    def runSimulation(self):

        # Setting Sumo Binary
        if self.options.nogui:
            sumoBinary = checkBinary("sumo")
        else:
            sumoBinary = checkBinary("sumo-gui")   

        # Running the loop
        traci.start([sumoBinary, "-c", self.pathCFG, "--tripinfo-output", self.outPath + "tripinfo.xml", "--time-to-teleport", "-1"])
        self.loop()
    



import os
import sys
import optparse
import traci
import sumolib
from sumolib import checkBinary 

"""
    @Class: Main class for emission analysis
"""
class EmissionJunction:

    """
        @Method: Initialising variables and config
    """
    def __init__(self, maxTimeSteps, pathCFG, outPath, pathNET, pathSummaryFile, numLocs, initialPoiLocation = (5064.74,3568.48)):
        
        # Check for Bash Shortcut
        if "SUMO_HOME" in os.environ:
            sys.path.append(os.path.join(
                os.environ["SUMO_HOME"],
                "tools"
            ))
        else:
            sys.exit("Please populate $SUMO_HOME in your .bashrc file.") 

        # Assign all variables
        self.maxTimeSteps = maxTimeSteps
        self.pathCFG = pathCFG 
        self.outPath = outPath 
        self.pathNET = pathNET
        self.pathSummaryFile = pathSummaryFile
        self.numLocs = numLocs
        self.initialPoiLocation = initialPoiLocation

        # Option Parser
        optParser = optparse.OptionParser()
        optParser.add_option(
            "--nogui", 
            action = "store_true",
            default = False,
            help = "Please run sumo --help."
        )
        options, args = optParser.parse_args()
        self.options = options

        # Initialise variables used in the simulation
        self.net = sumolib.net.readNet(pathNET)
        self.junctions = self.net.getNodes()
        self.junctionsEmissions = []

        for _ in range(len(self.junctions)):
            self.junctionsEmissions.append([])


    """
        @Method: Calculates the sum of CO2 emissions for each lane that pertains to the junction
    """
    def get_emission_at(self, junction):
        tc2e = 0
        incoming_edges = junction.getIncoming()
        outgoing_edges = junction.getOutgoing()
        edges = incoming_edges+outgoing_edges
        for edge in edges:
            lanes = edge.getLanes()
            for lane in lanes:
                lane_id = lane.getID()
                co2 = traci.lane.getCO2Emission(lane_id)
                tc2e+=co2
            
        return tc2e


    """
        @Method: TraCI simulation step and calculation of positions - run in the runSimulation() method
    """
    def loop(self):
        traci.poi.add("index", self.initialPoiLocation[0], self.initialPoiLocation[1], (255,255,255), layer=202.0)
        step = 0

        junctions = self.junctions 
        locs = min(5, self.numLocs)

        junctions_emissions = []
        for i in range(len(junctions)):
            junctions_emissions.append([])

        steps = []
        max_list = []
        max_dict = {}

        while traci.simulation.getMinExpectedNumber() > 0:

            traci.simulationStep()
            traci.poi.setType('index', 'TimeStep = '+str(step))
            
            if step%100 == 0:
                if(step!=0):
                    for i in range(locs):
                        traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
                max_list = []
                max_dict = {}
                
                for idx, junction in enumerate(junctions):
                    emission = self.get_emission_at(junction)
                    junctions_emissions[idx].append(emission)
                    max_dict[idx] = emission

                max_list = sorted(max_dict.items(), key=lambda x:x[1])
                max_list.reverse()

                steps.append(step)
                for i in range(locs):
                    traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
                
                filename = self.outPath+'screenshots/00'+str(step)+'.jpg'
                traci.gui.screenshot('View #0', filename)
            step += 1
        
        # Printing the values found
        traci.poi.setType('index', 'Aggregate')

        total_emissions_dict = {}
        for idx, emission in enumerate(junctions_emissions):
            total_emissions_dict[idx] = sum(emission)
        total_emissions_list = sorted(total_emissions_dict.items(), key= lambda x:x[1])
        total_emissions_list.reverse()

        for i in range(locs):
            traci.gui.toggleSelection(junctions[max_list[i][0]].getID(), objType='junction')
        
        f = open(self.pathSummaryFile, 'w')
        for i in range(locs):
            traci.gui.toggleSelection(junctions[total_emissions_list[i][0]].getID(), objType='junction')
            
            x,y = traci.junction.getPosition(junctions[total_emissions_list[i][0]].getID())
            lon, lat = traci.simulation.convertGeo(x, y)
            lon, lat = self.net.convertXY2LonLat(x, y)
            
            f.write(str(junctions[total_emissions_list[i][0]].getID())+ +" Coordinates:" + str(lat) + " " + str(lon) +  '\n')
        f.close()

        filename = self.outPath+'screenshots/aggregate.jpg'
        traci.simulationStep()
        traci.gui.screenshot('View #0', filename)    

        traci.simulationStep()
        traci.close()


    """
        @Method: Main Method for Speed Camera simulation
    """
    def runSimulation(self):

        # Setting Sumo Binary
        if self.options.nogui:
            sumoBinary = checkBinary("sumo")
        else:
            sumoBinary = checkBinary("sumo-gui")   

        # Running the loop
        traci.start([sumoBinary, "-c", self.pathCFG, "--tripinfo-output", self.outPath + "tripinfo.xml", "--time-to-teleport", "-1"])
        self.loop()  



import math
import os
import sys
import optparse
import traci
from sumolib import checkBinary 
from collections import Counter

"""
   @Class: Helper class to store vehicle attributes
"""
class Vehicle:
    _laneId = None 
    _speedLastTimeStep = None 

    def __init__(self, laneId, speedLastTimeStep) -> None:
        self._laneId = laneId
        self._speedLastTimeStep = speedLastTimeStep

    def getLaneId(self) -> int:
        return self._laneId

    def getSpeedLastTimeStep(self) -> float:
        return self._speedLastTimeStep

    @staticmethod
    def checkPositionInBox(left, right, up, down, x, y):
        if x < left or x > right:
            return False 
        if y < down or y > up:
            return False 
        return True

"""
    @Class: Helper class to store vehicle position and speed values
"""
class VehiclePosition:
        _x = None 
        _y = None 
        _speedLastTimeStep = None 

        def __init__(self, x, y, speedLastTimeStep) -> None:
            self._x = x 
            self._y = y 
            self._speedLastTimeStep = speedLastTimeStep

        def getPosition(self):
            return self._x, self._y
        
        def getSpeedLastTimeStep(self) -> float:
            return self._speedLastTimeStep

"""
    @Class: Data Structure to store speed and position tuples
"""
class SpeedPositionHelper:
    _positionList = None
    _speed = None 

    def __init__(self, positionList, speed) -> None:
        self._positionList = positionList
        self._speed = speed
    
    def getPosition(self) -> list:
        return self._positionList

    def getSpeed(self) -> float:
        return self._speed
    
    def getX(self) -> float:
        return self._positionList[0]

    def getY(self) -> float:
        return self._positionList[1]

"""
    @Class: Distance calculation and nearest neighbour removal functions
"""
class DistanceCalculationHelper:
    _nearest = None 
    _pointList = None

    def __init__(self, nearest, sortedPointList):
        self._nearest = nearest
        self._pointList = sortedPointList
    
    @staticmethod
    def getDistanceBetweenPoints(x1, x2, y1, y2):
        return math.sqrt(
            (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)
        )
    
    def removeNearestList(self):
        l = []
        i = 0 
        while i < len(self._pointList):
            l.append(self._pointList[i])
            j = (i+1)
            while j < len(self._pointList):
                x1 = self._pointList[i][0]
                y1 = self._pointList[i][1]
                x2 = self._pointList[j][0]
                y2 = self._pointList[j][1]

                d = DistanceCalculationHelper.getDistanceBetweenPoints(x1,x2,y1,y2)
                if d <= self._nearest:
                    j += 1 
                else:
                    break
            i = j    
        return l 
        
"""
    @Class: Main class to find speed camera locations
"""
class SpeedCamera:
    """
        @Method: Initialing variables and config
    """
    def __init__(self, maxTimeSteps, nearestNeighbourDisallow, gridArray, pathCFG, outPath, summaryFilePath, numLocs):
        
        # Check for Bash Shortcut
        if "SUMO_HOME" in os.environ:
            sys.path.append(os.path.join(
                os.environ["SUMO_HOME"],
                "tools"
            ))
        else:
            sys.exit("Please populate $SUMO_HOME in your .bashrc file.") 

        # Assign all variables
        self.maxTimeSteps = maxTimeSteps
        self.nearestNeighbourDisallow = nearestNeighbourDisallow
        self.left = gridArray[0]
        self.right = gridArray[1]
        self.up = gridArray[2]
        self.down = gridArray[3]
        self.pathCFG = pathCFG
        self.outPath = outPath
        self.summaryFilePath = summaryFilePath
        self.numLocs = numLocs

        # Option Parser
        optParser = optparse.OptionParser()
        optParser.add_option(
            "--nogui", 
            action = "store_true",
            default = False,
            help = "Please run sumo --help."
        )
        options, args = optParser.parse_args()
        self.options = options

        # Variables used in the simulation
        self.vehiclesDict = {}
        self.freqDict = {}
        self.iters = 0 
        self.previousPOI = []
        self.deleteOldPOI = False 


    """
        @Method: Helper function for sorting based on speed values
    """
    def speedCompareKey(self, item):
        return -item.getSpeed()
    
    """
        @Method: TraCI simulation step and calculation of positions - run in the runSimulation() method
    """
    def loop(self):
        timeStep = 0 

        while traci.simulation.getMinExpectedNumber() > 0 and timeStep <= self.maxTimeSteps:
            traci.simulationStep()

            positionsList = []
            for v in traci.vehicle.getIDList():
                x,y = traci.vehicle.getPosition(v) 
                cs = traci.vehicle.getSpeedWithoutTraCI(v)
                
                if Vehicle.checkPositionInBox(self.left,self.right,self.up,self.down,x,y):
                    if v not in self.vehiclesDict or cs > self.vehiclesDict[v].getSpeedLastTimeStep():
                        vo = VehiclePosition(x = x, y = y, speedLastTimeStep = cs)
                        self.vehiclesDict[v] = vo 
                            
                        xys = SpeedPositionHelper(positionList = [x,y], speed = cs)
                        positionsList.append(xys)
                            
                        if (x,y) not in self.freqDict: 
                            self.freqDict[(x,y)] = 1 
                        else:
                            self.freqDict[(x,y)] += 1

            positionsList.sort(key = self.speedCompareKey)

            # Closest Distance Logic
            distList = []
            for p in range(len(positionsList)):
                l = []
                l.append(positionsList[p].getX())
                l.append(positionsList[p].getY())
                distList.append(l)

            dcHelper = DistanceCalculationHelper(self.nearestNeighbourDisallow, distList)
            positionsList = dcHelper.removeNearestList()
            
            for i in range(min(5, len(positionsList))):
                nid = "demo" + str(self.iters)
                self.iters += 1

                if self.iters % 200 == 0:
                    if self.deleteOldPOI == True:
                        for oldID in self.previousPOI:
                            try:
                                traci.poi.remove(oldID)
                            except Exception as e:
                                print(e)
                    self.previousPOI = []
                    traci.poi.add(poiID=nid, x = positionsList[i][0], y = positionsList[i][1], layer = 202.0, color = (255,0,0))
                    traci.gui.toggleSelection(nid, objType='poi')
                    self.previousPOI.append(nid)

            timeStep += 1 

    """
        @Method: Printing Location Info
    """
    def printLocations(self):
        traci.close()
        sys.stdout.flush()
        print("-"*100)
        numLocs = self.numLocs
        numLocs = min(numLocs, len(self.freqDict.keys()))
        c = Counter(self.freqDict)
        for xyloc in c.most_common()[:numLocs]:
            print(f"Position, Violation count: {xyloc}")

        with open(self.summaryFilePath,'w') as f:
            for xyloc in c.most_common()[:numLocs]:
                f.write(f"Position, Violation count: {xyloc}" + "\n")
    

    """
        @Method: Main Method for Speed Camera simulation
    """
    def runSimulation(self):

        # Setting Sumo Binary
        if self.options.nogui:
            sumoBinary = checkBinary("sumo")
        else:
            sumoBinary = checkBinary("sumo-gui")   

        # Running the loop
        traci.start([sumoBinary, "-c", self.pathCFG, "--tripinfo-output", "tripinfo.xml", "--time-to-teleport", "-1"])
        self.loop()

        # Printing the locations and saving screenshots
        filename = self.outPath + 'speedcams.jpg'
        traci.simulationStep()
        traci.gui.screenshot('View #0', filename)    
        traci.simulationStep()
        self.printLocations()



import xml.etree.ElementTree as ET

"""
    @Class : To perform edge change interventions
"""

class ChangeEdges():
    def __init__(self, fileName):
        self._fileName = fileName
        self._tree = ET.parse(self._fileName)
        self._root = self._tree.getroot()
    
    """
        Get Head of XML Tree
    """
    def getRootElementTag(self):
        return self._root.tag
    
    """
        Get different attribute types
    """
    def getUniqueParentTags(self):
        d = {}
        for child in self._root:
            if (child.tag) not in d:
                d[child.tag] = 1
            else:
                d[child.tag] += 1

        return d.keys()

    """
        Get all Edge IDs
    """
    def getUniqueEdgeIDs(self):
        l = []
        for child in self._root:
            if child.tag == "edge":
                l.append(child.attrib["id"])
        
        return list(set(l))

    """
        Get Edge Information
    """
    def getEdgeInformation(self, edgeIdList):
        for child in self._root:
            if child.tag == "edge":
                if child.attrib["id"] in edgeIdList:
                    for lanes in child:
                        print(lanes.attrib)

    """
        Disallow certain vehicle types at edges
        @Link: https://sumo.dlr.de/docs/Vehicle_Type_Parameter_Defaults.html
    """
    def disallowAppendTypes(self, vehicleTypes, edgeIdList, newFileName = None):
        for child in self._root:
            if child.tag == "edge":
                if child.attrib["id"] in edgeIdList:
                    for lanes in child:
                        attributesDict = lanes.attrib
                        if "disallow" in attributesDict:
                            s = attributesDict["disallow"]
                            for v in vehicleTypes:
                                if v not in s:
                                    s += " "
                                    s += str(v)
                            lanes.set("disallow", s)

        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))

    """
        Allow certain vehicle types at edges
        @Link: https://sumo.dlr.de/docs/Vehicle_Type_Parameter_Defaults.html
    """
    def allowAppendTypes(self, vehicleTypes, edgeIdList, newFileName = None):
        for child in self._root:
            if child.tag == "edge":
                if child.attrib["id"] in edgeIdList:
                    for lanes in child:
                        attributesDict = lanes.attrib
                        if "allow" in attributesDict:
                            s = attributesDict["allow"]
                            for v in vehicleTypes:
                                if v not in s:
                                    s += " "
                                    s += str(v)
                            lanes.set("allow", s)

        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))


import xml.etree.ElementTree as ET

"""
    @Class : To perform trip interventions
"""
class ChangeTrips():
    def __init__(self, fileName):
        self._fileName = fileName
        self._tree = ET.parse(self._fileName)
        self._root = self._tree.getroot()

    """
        Get Head of XML Tree
    """
    def getRootElementTag(self):
        return self._root.tag

    """
        Get different attribute types
    """
    def getUniqueParentTags(self):
        d = {}
        for child in self._root:
            if (child.tag) not in d:
                d[child.tag] = 1
            else:
                d[child.tag] += 1

        return d.keys()

    """
        Get all Trip IDs
    """
    def getLaneTypes(self):
        l = []
        for child in self._root:
            if child.tag == "trip":
                l.append(child.attrib["id"])

        return l

    """
        Change the start time of a trip
    """
    def changeTripStartTime(self, tripIDList, newStartTime, newFileName = None):
        for child in self._root:
            if child.tag == "trip" and newStartTime >= 0:
                attributesDict = child.attrib
                if int(attributesDict["id"]) in tripIDList:
                    child.set("depart", str(newStartTime))

        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))

    """
        Get Trip information
    """
    def getTripInformation(self):
        l = []
        for child in self._root:
            if child.tag == "trip":
                l.append(
                    (child.attrib["id"],
                    child.attrib)
                )
        return l


import xml.etree.ElementTree as ET

"""
    @Class : To perform lane change interventions
"""
class ChangeLanes():
    def __init__(self, fileName):
        self._fileName = fileName
        self._tree = ET.parse(self._fileName)
        self._root = self._tree.getroot()

    """
        Get Head of XML Tree
    """
    def getRootElementTag(self):
        return self._root.tag

    """
        Get different attribute types
    """
    def getUniqueParentTags(self):
        d = {}
        for child in self._root:
            if (child.tag) not in d:
                d[child.tag] = 1
            else:
                d[child.tag] += 1

        return d.keys()

    """
        Get all Lane IDs
    """
    def getLaneTypes(self):
        l = []
        for child in self._root:
            if child.tag == "type":
                l.append(child.attrib["id"])

        return l

    """
        Change priority attributes of certain lanes
    """
    def changePriorityLanes(self, laneTypeList, newPriorityValue, newFileName = None):
        for child in self._root:
            if child.tag == "type":
                attributesDict = child.attrib
                if attributesDict["id"] in laneTypeList:
                    child.set("priority", str(newPriorityValue))
        
        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))

    """
        Change the number of lanes with given IDs
    """
    def changeNumLanes(self, laneTypeList, newNumberLanes, newFileName = None):
        for child in self._root:
            if child.tag == "type":
                attributesDict = child.attrib
                if attributesDict["id"] in laneTypeList:
                    child.set("numLanes", str(newNumberLanes))

        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))

    """
        Make certain lanes one-way, or reverse their states
    """
    def toggleOneWay(self, laneTypeList, newFileName = None):
        for child in self._root:
            if child.tag == "type":
                attributesDict = child.attrib
                if attributesDict["id"] in laneTypeList:
                    if str(attributesDict["oneway"]) == "0":
                        child.set("oneway", "1")
                    elif str(attributesDict["oneway"]) == "1":
                        child.set("oneway", "0")
        if newFileName is None:
            self._tree.write(self._fileName)
        else:
            self._tree.write(str(newFileName))

    """
        Get Lane information
    """
    def getLaneInformation(self):
        l = []
        for child in self._root:
            if child.tag == "type":
                l.append(
                    (child.attrib["id"],
                    child.attrib)
                )
        return l